//package syntax.ast;

import java_cup.runtime.ComplexSymbolFactory.Location;

/** La classe abstraite ancêtre pour l'AST.
 * <ul><li>Construction VarArgs
 * <li>Itérable : {@code for (AstNode fils : pere) }
 * <li>Impression de l'arbre : {@link #toPrint}
 * <li>Gestion des positions dans le source : {@link #addPosition} </ul> */
public abstract class AstNode implements Iterable<AstNode> {
  /** Un nom pour le debugging. Par défaut le nom de la classe instanciante */
  private final String label;

  /** La liste des nœuds fils. */
  private final java.util.List<AstNode> fils;

  /** La localisation dans le fichier source: Debut. */
  private Location start;

  /** La localisation dans le fichier source: Fin. */
  private Location stop;

  /** Construit le nœud avec ses fils en VarArgs.
     @param fils une liste quelconque de fils */
  protected AstNode(final AstNode... fils) {
    this.label = getClass().getSimpleName();
    this.fils = new java.util.ArrayList<>();
    for (AstNode f : fils) this.fils.add(f);
  }

  /** Construit les fils par itération, uniquement pour {@link AstList}.
   * @param f le nœud fils à ajouter*/
  protected void addFils(final AstNode f) {
    this.fils.add(f);
  }

  /** Réalise le patron de conception "Visitor".
   * @param v le visiteur */
  public abstract void accept(AstVisitor v);

  /** Itération sur les fils.
   *  <br>{@code AstNode p; for (AstNode f : p) ...} */
  public java.util.Iterator<AstNode> iterator() {
    return fils.iterator();
  }

  /** Donne la taille.
   * @return le nombre de fils */
  public int size() { return this.fils.size(); }

  /** Gestion optionnelle des positions des symboles dans le source.
   * <br>Exemple dans CUP :
   * <pre>x ::= A:a .. Z:z {: RESULT = new Node(A,..,Z);
   *                    RESULT.addPosition(axleft,zxright); :}</pre>
   * @param left la position de début du symbole
   * @param right la position de fin du symbole */
  public void addPosition(final Location left, final Location right) {
    this.start = left;
    this.stop = right;
  }

  /** Imprime un noeud. */
  public String toString() {
    if (start == null && stop == null) return label;
    return label + "[" + locStr(start) + "-" + locStr(stop) + "]";
  }

  /** "Redéfinition" de {@code Location.toString()}.
   * @param l la Location à imprimer
   * @return la chaîne imprimable */
  private String locStr(final Location l) {
    if (l == null) return "";
    return l.getLine() + "/" + l.getColumn() + "(" + l.getOffset() + ")";
  }

  /** Imprime l'Arbre de Syntaxe Abstraite.
   * @return la représentation textuelle de L'AST */
  public String toPrint() {
    final StringBuilder sb = new StringBuilder();
    Print.astPrint(this, "", sb);
    return sb.toString();
  }

  /** Impression récursive de l'AST. */
  private static final class Print {
    /** Les jeux de caractères pour impression d'arbre.
     * <ul><li> ASCII : "|-", "| ", "\\-", "  "
     * <li> Unicode: "\u251c\u2500","\u2502\u0020","\u2514\u2500","\u0020\u0020"
     * <li> Unicode Compact: "\u251c", "\u2502", "\u2514", "\u0020"</ul> */
    private static final String[][] CSS = {
      {"|-", "| ", "\\-", "  "}, // ASCII
      {"├─", "│ ", "└─", "  "},  // Unicode
      {"├", "│", "└", " "} // Unicode Compact
    };
    /** Le jeu de caractères pour impression d'arbre. */
    private static final String[] CS = CSS[1]; // Unicode
    /** Constructeur pour la forme. */
    private Print() { }

    /** Imprime récursivement un arbre à la ASCII art.
     * @param indent la chaîne d'indentation courante */
    private static void astPrint(final AstNode n, final String indent,
        final StringBuilder sb) {
      sb.append(n + System.lineSeparator());
      for (java.util.Iterator<AstNode> it = n.iterator(); it.hasNext();) {
        final AstNode f = it.next();
        final int indice = it.hasNext() ?  0 : 2; /* last or not?*/
        sb.append(indent + CS[indice]);
        astPrint(f, indent + CS[indice + 1], sb);
      }
    }
  }
}
